import argparse
import pathlib
import collections
import sys
from typing import Union

import yaml
from tqdm.autonotebook import tqdm

from . import question
from . import remote
from . import gift
from . import colors


def main():

	parser = argparse.ArgumentParser(description='Build GIFT files (Moodle) from a simple specification')

	parser.add_argument(
		'-p', '--parameters_file', type=argparse.FileType('r'), default='parameters.yaml', help='parameters file',
		nargs='?')

	parser.add_argument(
		'-i', '--input_file', type=argparse.FileType('r'), default='bank.yaml', help='questions file', nargs='?')

	parser.add_argument(
		'-l', '--local', default=False, action='store_true', help="don't try to copy the images to the server")

	parser.add_argument(
		'-n', '--no-checks', default=False, action='store_true', help="don't check latex formulas (much faster)")

	parser.add_argument(
		'-o', '--overwrite-existing-latex-files', default=False, action='store_true',
		help="overwrite existing latex files if necessary instead of exiting")

	parser.add_argument(
		'-e', '--embed-images', default=False, action='store_true',
		help="embed the images rather than link to them (experimental)")

	command_line_arguments = parser.parse_args()

	wrap(
		parameters=command_line_arguments.parameters_file.name,
		questions_file=command_line_arguments.input_file.name, local_run=command_line_arguments.local,
		no_checks=command_line_arguments.no_checks,
		overwrite_existing_latex_files=command_line_arguments.overwrite_existing_latex_files,
		embed_images=command_line_arguments.embed_images)


def wrap(
		parameters: str, questions_file: str, local_run: bool, no_checks: bool,
		overwrite_existing_latex_files: bool, embed_images: bool):

	# ================================= parameters' reading

	# if a file name was passed, either as a string or wrapped in a `Pathlib`,...
	if (type(parameters) == str) or (type(parameters) == pathlib.Path):

		# ...it is read
		with open(parameters) as yaml_data:

			parameters = yaml.load(yaml_data, Loader=yaml.FullLoader)

	# if a file name was *not* passed...
	else:

		# ...then it should be a dictionary
		assert type(parameters) == dict

	input_file = pathlib.Path(questions_file)

	# ================================= questions' reading

	# the file containing the questions is read
	with open(input_file) as yaml_data:

		input_data = yaml.load(yaml_data, Loader=yaml.FullLoader)

	categories = input_data['categories']
	pictures_base_directory = input_data['pictures base directory']

	# =================================

	# if images embedding was requested...
	if embed_images:

		# ...a connection is not needed
		connection = None

	# if images are *not* to be embedded...
	else:

		# if "local" running was requested...
		if local_run:

			# ...a "fake" connections is instantiated
			connection = remote.FakeConnection(parameters['images hosting']['copy']['host'])

		# if NO local running was requested...
		else:

			# ...an actual connection with the requested host is opened
			connection = remote.Connection(
				parameters['images hosting']['copy']['host'], **parameters['images hosting']['ssh'])

	# output file has the same name as the input with the ".gift.txt" suffix
	output_file = input_file.with_suffix('.gift.txt')

	# to keep track of files already compiled/transferred
	history = {'already compiled': set(), 'already transferred': set()}

	latex_auxiliary_file = pathlib.Path(parameters['latex']['auxiliary file'])

	# if overwriting files was not requested AND latex checks are enabled AND the given auxiliary file already exists...
	if (not overwrite_existing_latex_files) and (not no_checks) and latex_auxiliary_file.exists():

		print(
			f'{colors.error}latex auxiliary file {colors.reset}"{latex_auxiliary_file}"{colors.error} exists'
			f' and would be overwritten (pass "-o" to skip this check)')
		sys.exit(1)

	with open(output_file, 'w') as f:

		# for every category...
		for cat in tqdm(categories, desc='category'):

			# if "something" was actually provided...
			if cat['name']:

				# ...if it's *not* a list...
				if type(cat['name']) != list:

					# ...it is turned into one
					cat['name'] = [cat['name']]

				for c in cat['name']:

					f.write(gift.from_category(c))

			# the names of *all* the questions
			all_names = [q['name'] for q in cat['questions']]

			# list with names that show up more than once
			duplicates = [name for name, count in collections.Counter(all_names).items() if count > 1]

			# all the names should be different
			assert not duplicates, f'{colors.error}duplicates in category {colors.reset}{cat["name"]}: {duplicates}'

			# for every question in the category...
			for q in tqdm(cat['questions'], desc='question', leave=True):

				# the class that will be instantiated for this particular question; notice the class is removed
				# (popped) from the dictionary so that it doesn't get passed to the initializer
				question_class = getattr(question, q.pop('class'))

				# whether or not latex formulas should be checked...
				q['check_latex_formulas'] = not no_checks

				# ...if so, this auxiliary file will be used (created)
				q['latex_auxiliary_file'] = latex_auxiliary_file

				# "history" is passed
				q['history'] = history

				if embed_images:

					# question is instantiated and "decorated"
					q = question.SvgToInline(question.TexToSvg(question_class(**q)))

				else:

					# question is instantiated and "decorated"
					q = question.SvgToHttp(
						question.TexToSvg(
							question_class(**q)
						), connection, parameters['images hosting']['copy']['public filesystem root'],
						pictures_base_directory, parameters['images hosting']['public URL'])

				f.write(f'{q.gift}\n\n')

	print(f'{colors.info}file "{colors.reset}{output_file}{colors.info}" created')

	# if images are *not* to be embedded...
	if not embed_images:

		# if this is a "fake" connection
		if type(connection) == remote.FakeConnection:

			# if there is any file to be copied...
			if connection.files_to_copy:

				print(f'{colors.info}you *should* copy:')

				for source, remote_directory in connection.files_to_copy:

					print(
						f'{source}{colors.info} to '
						f'{colors.reset}{remote_directory}{colors.info} in {colors.reset}{connection.host}')

	# if latex checks are enabled AND some formula was processed...
	if (not no_checks) and latex_auxiliary_file.exists():

		# latex auxiliary files are deleted
		for suffix in ['.tex', '.aux', '.log']:

			latex_auxiliary_file.with_suffix(suffix).unlink()



import sys
import abc
import functools
import re
import pathlib
import string
from typing import Union, Callable, Optional

from . import gift
from . import image
from . import remote
from . import colors
from . import latex

# regular expression to extract a percentage
re_percentage = re.compile(r'(\d*\.\d+|\d+)\s*%')


def markdown_header(
		text: str,
		template: string.Template = string.Template('<span style="font-family:Papyrus; font-size:2em;">$text</span>')
) -> str:
	"""
	Returns a markdown header for a given text.

	Parameters
	----------
	text : str
		Text for the header.
	template : string template
		The template that defines the appearance of the header, and in which `text` will be embedded.

	Returns
	-------
	str:
		Markdown-compatible text.

	"""

	return f'\n{template.substitute(text=text)}\n\n'


class HtmlQuestion(metaclass=abc.ABCMeta):
	"""
	Abstract class implementing an html-based question.
	"""

	# "\" is duplicated because it is assumed it has been escaped previously by `gift.process_latex`
	latex_in_text_substitutions = [
		# \textbf
		[r'\\textbf{([^}]+)}', r'<b>\1</b>', r'<b>([^<]*)</b>', r'\\textbf{\1}'],
		# \textit
		[r'\\textit{([^}]+)}', r'<i>\1</i>', r'<i>([^<]*)</i>', r'\\textit{\1}']
	]

	def __init__(
			self, name: str, statement: str, history: dict, check_latex_formulas: bool,
			latex_auxiliary_file: Union[str, pathlib.Path], images_settings: Optional[dict] = None,
			feedback: Optional[str] = None, time: Optional[int] = None):
		"""
		Initializer.

		Parameters
		----------
		name : str
			Name of the question.
		statement : str
			Statement of the question.
		history: dict
			Variable storing a "context" common to all the questions.
		check_latex_formulas: bool
			Whether or not an attempt should be made to compile every latex formula to detect errors.
		latex_auxiliary_file: str or Pathlib
			The latex file that will be created to check the formulas.
		images_settings : dict, optional
			width and height of *all* the images in the question.
		feedback : str, optional
			Feedback for the question.
		time: int, optional
			The number of minutes deemed necessary to answer the question.
		"""

		self.name = name
		self.statement = statement.rstrip()
		self.feedback = feedback
		self.time = time
		self.history = history

		if images_settings is None:

			self.images_width, self.images_height = None, None

		else:

			assert ('width' in images_settings) and ('height' in images_settings), \
				'"width" and/or "height" missing in "image_settings"'

			self.images_width, self.images_height = images_settings['width'], images_settings['height']

		self.processing_functions = [
			functools.partial(gift.process_url_images, width=self.images_width, height=self.images_height),
			gift.process_new_lines, functools.partial(
				gift.process_latex, latex_auxiliary_file=latex_auxiliary_file, check_compliance=check_latex_formulas)
		]

		# functions to process LaTeX commands *in text* (ignoring occurrences in formulas)...
		latex_text_processing_functions = [
			functools.partial(latex.replace_and_replace_only_in_formulas, *l) for l in self.latex_in_text_substitutions]

		# ...are added to the pool
		self.processing_functions += latex_text_processing_functions

		# these might be tampered with by subclasses/decorators
		self.pre_processing_functions = []
		self.post_processing_functions = []

	def process_text(self, text: str) -> str:
		"""
		Functions in `self.processing_functions` are applied on the given input.

		Parameters
		----------
		text : str
			Input text.

		Returns
		-------
		out: str
			Text processed with the previously defined functions.

		"""

		for function in (self.pre_processing_functions + self.processing_functions + self.post_processing_functions):

			try:

				text = function(text)

			except gift.NotCompliantLatexFormula as e:

				print(
					f'\n{colors.error}cannot compile latex formula\n {colors.extra_info}{e.formula}{colors.reset} in '
					f'{colors.info}{self.name}')

				sys.exit(1)

		return text

	@property
	def gift(self):
		"""
		Builds the question in the GIFT format.

		Returns
		-------
		out: str
			GIFT-ready text.

		"""

		# feedback part of the answer if any
		feedback = ("\n\t" + gift.from_feedback(self.process_text(self.feedback.rstrip()))) if self.feedback else ""

		# the full answer
		answer = f'{{\n{self.answer + feedback}\n}}'

		# if a `time` estimate was passed...
		if self.time:

			# ...it is appended at the end of the statement
			self.statement += '\n\n\n' + r'<i>Estimated time\: ' + str(self.time) + r' minutes</i>' + '\n'

		return gift.from_question_name(self.name) + gift.html + self.process_text(self.statement) + answer

	def __repr__(self):

		return f'Name: {self.name}\nStatement: {self.statement}'

	@property
	@abc.abstractmethod
	def answer(self):
		"""
		Abstract method yielding the GIFT-ready answer to the question.

		Returns
		-------
		out: str
			GIFT-ready text with the answer.

		"""

		pass

	def to_jupyter(self):

		statement = self.statement

		# if a `time` estimate was passed...
		if self.time:

			# ...it is appended at the end of the statement
			statement += f'\n\n\n*Estimated time: {self.time} minutes*\n'

		feedback = (f'{markdown_header("Feedback")}' + self.feedback.rstrip()) if self.feedback else ''

		# a copy of each list is made so that the class attribute is not modified
		latex_in_text_substitutions = [e.copy() for e in self.latex_in_text_substitutions[:2]]

		# \textbf-related substitutions are tweaked
		latex_in_text_substitutions[0][1] = r'**\1**'
		latex_in_text_substitutions[0][2] = r'\*\*([^\*]+)\*\*'

		# \textit-related substitutions are tweaked
		latex_in_text_substitutions[1][1] = r'*\1*'
		latex_in_text_substitutions[1][2] = r'\*([^\*]+)\*'

		# just like LaTeX, in markdown a single "newline" doesn't reflect in the output, and hence every new line
		# is duplicated
		latex_in_text_substitutions += [['\n', '\n\n', '\n\n', '\n']]

		def apply_substitutions(text: str, substitutions: list):

			res = text

			for s in substitutions:

				res = latex.replace_and_replace_only_in_formulas(*s, res)

			return res

		statement = apply_substitutions(statement, latex_in_text_substitutions)
		feedback = apply_substitutions(feedback, latex_in_text_substitutions)

		return f'{markdown_header("Statement")}{statement}\n{feedback}\n'


class Numerical(HtmlQuestion):
	"""
	Class implementing a numerical question.
	"""

	def __init__(
			self, name: str, statement: str, history: dict, check_latex_formulas: bool,
			latex_auxiliary_file: Union[str, pathlib.Path], solution: dict, images_settings: Optional[dict] = None,
			feedback: Optional[str] = None, time: Optional[int] = None):
		"""
		Initializer.

		Parameters
		----------
		solution : dict
			Value and, optionally, tolerated error of the solution.
		"""

		super().__init__(
			name, statement, history, check_latex_formulas, latex_auxiliary_file, images_settings, feedback, time)

		assert ('value' in solution), '"value" missing in "solution"'

		self.solution_value = str(solution['value'])

		if 'error' in solution:

			self.solution_error = ':'

			# try to match a percentage
			m = re_percentage.match(str(solution['error']))

			# if so...
			if m:

				self.solution_error += str(solution['value'] * float(m.group(1)) / 100.)

			else:

				self.solution_error += str(solution['error'])

		# an error was NOT provided
		else:

			self.solution_error = ''

	def __repr__(self):

		error = f' (error: {self.solution_error[1:]})' if self.solution_error != '' else ''

		return super().__repr__() + '\n' + f'Solution: {self.solution_value}' + error

	@property
	def answer(self):

		return '#\t=%100%' + self.solution_value + self.solution_error + '#'

	def to_jupyter(self):

		res = super().to_jupyter()

		return res + f'{markdown_header("Solution")} {self.solution_value} (error: {self.solution_error[1:]})\n'


class MultipleChoice(HtmlQuestion):
	"""
	Class implementing a multiple-choice question.
	"""

	template_wrong_answers = string.Template(r"**<font color='$color'>$text</font>**")

	def __init__(
			self, name: str, statement: str, history: dict, check_latex_formulas: bool,
			latex_auxiliary_file: Union[str, pathlib.Path], answers: dict, images_settings: Optional[dict] = None,
			feedback: Optional[str] = None, time: Optional[int] = None):
		"""
		Initializer.

		Parameters
		----------
		answers : dict
			Right answer and a list with the wrong ones.
		"""

		super().__init__(
			name, statement, history, check_latex_formulas, latex_auxiliary_file, images_settings, feedback, time)

		self.answers = answers

	@property
	def answer(self):

		processed_answers = []

		# the maximum grade allowed by partially correct answers
		max_grade = 0.

		for a in self.answers['wrong']:

			# if it is a list
			if isinstance(a, list):

				processed_answers.append(f'~%{a[1]}%{self.process_text(a[0])}')

				# if the grade is positive...
				if a[1] > 0:

					# ...it is accounted when computing the maximum
					max_grade += float(a[1])

			# if it is a scalar
			else:

				processed_answers.append(f'~{self.process_text(a)}')

		# if a "perfect" (100% credit) was provided...
		if 'perfect' in self.answers:

			# ...it is added in a special way
			processed_answers.insert(0, '=' + self.process_text(self.answers['perfect']))

		else:

			# if it's not possible to get full credit...
			if max_grade < 100.:
				print(
					f"{colors.extra_info}question{colors.reset} \"{self.name}\" {colors.extra_info} won't allow "
					f"full credit {colors.reset}{max_grade}")

		return '\t' + '\n\t'.join(processed_answers)

	def to_jupyter(self):

		res = super().to_jupyter()

		res += markdown_header('Choices')

		if 'perfect' in self.answers:

			# res += f'---\n'

			# res += f'* {self.answers["perfect"]}\n'
			res += f'* {self.template_wrong_answers.substitute(color="green", text=self.answers["perfect"])}\n'

			# res += f'---\n'

		for a in self.answers['wrong']:

			if isinstance(a, list):

				formatted_grade = self.template_wrong_answers.substitute(
					color='green' if float(a[1]) > 0 else 'red', text=f'{a[1]}%')

				res += f'* {a[0]} (**{formatted_grade}**)\n'
				# res += f'* {a[0]} (**{a[1]}%**)\n'

			else:

				# res += f'* {a}\n'
				res += f'* {self.template_wrong_answers.substitute(color="red", text=a)}\n'

		return res


# ========================================== Decorators

class QuestionDecorator:
	"""
	Abstract class to implement a question decorator.
	"""

	def __init__(self, decorated: Union[HtmlQuestion, 'QuestionDecorator']):

		self._decorated = decorated

	# any method/attribute not implemented here...
	def __getattr__(self, item):

		# ...is relayed to the *decorated* object
		return getattr(self._decorated, item)

	def __setattr__(self, key, value):

		# except for the `_decorated` attribute...
		if key == '_decorated':
			object.__setattr__(self, key, value)
		# ...everything else...
		else:
			# ...is relayed to the decorated object
			setattr(self._decorated, key, value)

	# TODO: when minimum Python version is forwarded to 3.8, `[re.Match]` should replace `...` as the the signature
	#  of the `Callable` in the type hinting for `replacement`
	@staticmethod
	def transform_files(
			text: str, pattern: str, process_match: Callable[[str], None],
			replacement: Union[str, Callable[..., str]]):
		"""
		It searches in a text for strings corresponding to files (maybe including a path), replaces them by another
		string according to some function and, additionally, processes each file according to another function.

		Parameters
		----------
		text : str
			Input text.
		pattern : str
			Regular expression including a capturing group that yields the file.
		process_match : Callable[[str], None]
			Function that will *process* each file.
		replacement : str or Callable[[re.Match], str]
			Regular expression making use of the capturing group or function processing the match delivered by `pattern`

		Returns
		-------
		out: str
			Output text with the replacements performed, after having processed all the files.

		"""

		# all the matching files in the given text
		files = re.findall(pattern, text)

		# breakpoint()

		# every one of them...
		for file in files:

			# ...is processed
			process_match(file)

		# replacement of matches
		return re.sub(pattern, replacement, text)


class TexToSvg(QuestionDecorator):
	"""
	Decorator to convert TeX files into svg files.
	"""

	def __init__(self, decorated: Union[HtmlQuestion, QuestionDecorator]):

		super().__init__(decorated)

		def process_match(f):

			# if this file has not been already compiled-converted...
			if f not in self.history['already compiled']:

				# ...it is...
				image.pdf_to_svg(image.tex_to_pdf(f))

				# ...and a note is made of it
				self.history['already compiled'].add(f)

			# else:
			#
			# 	print(f'{f} already compiled-converted...')

		# a new pre-processing function is attached to the corresponding list
		# (the "\1" in `replacement` refers to matches in `pattern`)
		# NOTE: an extra space is added in the replacement for `SvgToInline` --- not anymore
		self.pre_processing_functions.append(functools.partial(
			self.transform_files, pattern='(\S+)\.tex', process_match=process_match, replacement=r'\1.svg'))


class SvgToHttp(QuestionDecorator):
	"""
	Decorator to transfer svg files to a remote location.
	"""

	pattern_svg_file = r'(?<!\S)(?!http)(\S+\.svg)(?!\S)'

	def __init__(
			self, decorated: Union[HtmlQuestion, QuestionDecorator], connection: remote.Connection,
			public_filesystem_root: str, pictures_base_directory: str, public_url: str):

		super().__init__(decorated)

		# assembled remote path
		remote_subdirectory = pathlib.Path(public_filesystem_root).joinpath(pictures_base_directory)

		# TODO: when minimum Python version is forwarded to 3.8, `re.Match` should be the type hinting for `m`
		def replacement_function(m) -> str:

			file = pathlib.Path(m.group(0))

			return public_url + pictures_base_directory + '/' + file.as_posix()

		def process_match(f):

			# if this file has not been already transferred...
			if f not in self.history['already transferred']:

				# ...it is...
				connection.copy(f, remote_directory=remote_subdirectory / pathlib.Path(f).parent)

				# ...and a note is made of it
				self.history['already transferred'].add(f)

		# a new pre-processing function is attached to the corresponding list
		self.pre_processing_functions.append(functools.partial(
			self.transform_files, pattern=self.pattern_svg_file,
			process_match=process_match, replacement=replacement_function))


class SvgToMarkdown(QuestionDecorator):
	"""
	Decorator to reformat svg files for including them in markdown strings.
	"""

	def __init__(self, decorated: Union[HtmlQuestion, QuestionDecorator]):

		super().__init__(decorated)

		def process_match(f):

			pass

		# a new pre-processing function is attached to the corresponding list
		self.pre_processing_functions.append(functools.partial(
			self.transform_files, pattern='(\S+)\.svg',
			process_match=process_match, replacement=r'![](' + r'\1' + '.svg)'))


class SvgToInline(QuestionDecorator):
	"""
	Decorator to directly include svg files into a question.
	"""

	# notice the order is important: each pattern will result in a different post-processing function and they are
	# applied *sequentially*, each one on the result of the previous one
	patterns = [r'([\\/\w_\-]+\.svg)']

	def __init__(self, decorated: Union[HtmlQuestion, QuestionDecorator]):

		super().__init__(decorated)

		def process_match(f):

			pass

		# TODO: when minimum Python version is forwarded to 3.8, `re.Match` should be the type hinting for `m`
		def replacement_function(m) -> str:

			file = pathlib.Path(m.group(1))

			return image.svg_to_html(file)

		for p in self.patterns:

			# a new pre-processing function is attached to the corresponding list
			self.post_processing_functions.append(functools.partial(
				self.transform_files, pattern=p, process_match=process_match, replacement=replacement_function))



import pathlib
import shutil
import subprocess
import string
import re
from typing import Union, List, Optional


def compile_tex(
		source_file: Union[str, pathlib.Path], timeout: Optional[int], options: List[str] = ['halt-on-error']) -> int:
	"""
	Compiles a TeX file.

	Parameters
	----------
	source_file : str or pathlib.Path
		TeX file.
	timeout: int
		Seconds that are given to compile the source.
	options: list of str
		Options to be passed to `pdflatex`.

	Returns
	-------
	out: int
		The exit status of the call to `pdflatex`.

	"""

	source_file = pathlib.Path(source_file)

	path_to_compiler = shutil.which('pdflatex')

	assert path_to_compiler, 'cannot find pdflatex'

	command = [path_to_compiler] + [f'-{o}' for o in options] + [source_file.name]

	run_summary = subprocess.run(command, capture_output=True, cwd=source_file.parent, timeout=timeout)

	return run_summary.returncode


latex_template = string.Template(r'''
\documentclass{standalone}

\usepackage{amsmath}

\begin{document}

$$
$formula
$$

\end{document}
''')


def formula_can_be_compiled(formula: str, auxiliary_file: Union[str, pathlib.Path]) -> bool:
	"""
	Checks whether a latex formula can be compiled with the above template, `latex_template`.

	Parameters
	----------
	formula : str
		Latex formula.
	auxiliary_file : str or pathlib.Path
		(Auxiliary) TeX file that is created to check the formula.

	Returns
	-------
	out: bool
		`True` if the compilation finished with no errors.

	"""

	tex_source_code = latex_template.substitute(formula=formula)

	with open(auxiliary_file, 'w') as f:

		f.write(tex_source_code)

	exit_status = compile_tex(auxiliary_file, timeout=10, options=['halt-on-error', 'draftmode'])

	return exit_status == 0


def replace_and_replace_only_in_formulas(
		pattern: str, replacement: str, formula_pattern: str, formula_replacement: str, text: str) -> str:
	"""
	Replaces a matched expression but only if the match occurs *outside* a LaTeX formula.

	Notice that arguments `pattern` and `replacement`, on one hand, and `formula_pattern` and `formula_replacement`,
	on the other, are to be passed to Python's `re.sub`.

	Parameters
	----------
	pattern : str
		Regular expression to be matched globally.
	replacement : str
		Replacement string for `pattern`
	formula_pattern : str
		Regular expression to be matched in formulas.
	formula_replacement : str
		Replacement string for `formula_pattern`
	text : str
		String to be processed

	Returns
	-------
	out: str
		String with replacements.

	"""

	def process_formula(m) -> str:

		return re.sub(formula_pattern, formula_replacement, m.group(0))

	res = re.sub(pattern, replacement, text)
	res = re.sub(r'\$[^\$]*\$', process_formula, res)

	return res



import colorama

# color is reset at the end of every print
colorama.init(autoreset=True)

# ------ colors definitions

info = colorama.Fore.LIGHTWHITE_EX
reset = colorama.Style.RESET_ALL
error = colorama.Fore.RED
extra_info = colorama.Fore.LIGHTCYAN_EX



import re
import pathlib
from typing import Union

from . import latex


class NotCompliantLatexFormula(Exception):

	def __init__(self, formula: str) -> None:

		self.formula = formula

	def __str__(self) -> str:

		return self.formula


html = '[html]'


def from_question_name(name: str) -> str:
	"""
	Generates GIFT text for a question name.

	Parameters
	----------
	name : str
		Name of a question.

	Returns
	-------
	out: str
		GIFT-ready question name.

	"""

	return f'::{name}::'


def from_image_url(url: str, width: int, height: int) -> str:
	"""
	Generates GIFT text for a URL.

	Parameters
	----------
	url : str
		Input URL.
	width : int
		Desired width of the image.
	height : int
		Desired height of the image.

	Returns
	-------
	out: str
		GIFT-ready text.

	"""

	# either both are set or none
	assert not ((width is not None) ^ (height is not None)), 'both "width" and "height" should be set or none of them'

	# `width` is checked, but `height` could be as well used (at this point either both are set or none)
	if width is not None:

		# notice the space at the beginning
		width_height = f' width="{width}" height="{height}"'

	else:

		width_height = ''

	res = f'<img src="{url}" alt="" role="presentation" class="atto_image_button_text-bottom"{width_height}>'

	for to_be_escaped in [':', '~', '=']:

		res = res.replace(to_be_escaped, '\\' + to_be_escaped)

	return res


def from_category(name: str, within_the_course: bool = True) -> str:
	"""
	Generates GIFT text for a category.

	Parameters
	----------
	name : str
		Category name.
	within_the_course: bool
		If True, the category will belong to the course in which the questions are imported.

	Returns
	-------
	out: str
		GIFT-ready text.

	"""

	return '$CATEGORY: ' + (r'$course$/' if within_the_course else '') + name + '\n\n'


def from_feedback(text: str) -> str:
	"""
	Generates GIFT-ready text from a feedback string.

	Parameters
	----------
	text : str
		Input text.

	Returns
	-------
	out: str
		GIFT-ready text.

	"""

	return '#'*4 + text


def process_latex(text: str, latex_auxiliary_file: Union[str, pathlib.Path], check_compliance: bool = True) -> str:
	"""
	Adapts every occurrence of $$ to GIFT.

	Parameters
	----------
	text : str
		Input text.
	latex_auxiliary_file: str or pathlib.Path
		(Auxiliary) TeX file that is created to check the formula.
	check_compliance: bool
		Whether or not to check if the formula can be compiled.

	Returns
	-------
	out: str
		GIFT-ready text.

	"""

	# TODO: when minimum Python version is forwarded to 3.8, `re.Match` should be the type hinting for "m"
	def replacement(m) -> str:

		latex_source = m.group(1)

		if check_compliance:

			if not latex.formula_can_be_compiled(latex_source, auxiliary_file=latex_auxiliary_file):

				raise NotCompliantLatexFormula(latex_source)

		for to_be_escaped in ['\\', '{', '}', '=']:

			latex_source = latex_source.replace(to_be_escaped, '\\' + to_be_escaped)

		latex_source = latex_source.replace('&', '&amp;')

		return r'\\(' + latex_source + r'\\)'

	# it looks for strings between $'s (that do not include $ itself) and wraps them in \( and \)
	return re.sub(r'\$([^\$]*)\$', replacement, text)


def process_url_images(text: str, width: int, height: int) -> str:
	"""
	Adapts to GIFT every URL in a given text.

	Parameters
	----------
	text : str
		Input text.
	width : int
		Desired width of the image.
	height : int
		Desired height of the image.

	Returns
	-------
	out: str
		GIFT-ready text.

	"""

	return re.sub(
		'http(\S+)(?!\S)',
		lambda m: '<p>' + from_image_url(m.group(0), width=width, height=height) + '<br></p>',
		text)


def process_new_lines(text: str) -> str:
	"""
	Adapts every new line to GIFT.

	Parameters
	----------
	text : str
		Input text.

	Returns
	-------
	out: str
		GIFT-ready text.

	"""

	return text.replace('\n', '<br>')



import sys
import pathlib
import shutil
import subprocess
from typing import Union

from . import colors
from . import latex


def tex_to_pdf(source_file: Union[str, pathlib.Path], timeout: int = 10) -> pathlib.Path:
	"""
	Turns a TeX file into a pdf.

	Parameters
	----------
	source_file : str or pathlib.Path
		TeX file.
	timeout: int
		Seconds that are given to compile the source.

	Returns
	-------
	out: pathlib.Path
		The path to the compiled pdf.

	"""

	source_file = pathlib.Path(source_file)

	try:

		exit_status = latex.compile_tex(source_file, timeout=timeout)

	except subprocess.TimeoutExpired:

		print(
			f'{colors.error}could not compile {colors.reset}{source_file + ".tex"}'
			f' {colors.error}in {colors.reset}{timeout}{colors.error} seconds'
		)

		sys.exit(1)

	assert exit_status == 0, f'{colors.error}errors were found while compiling {colors.reset}{source_file}'

	return source_file.with_suffix('.pdf')


def pdf_to_svg(input_file: Union[str, pathlib.Path]) -> pathlib.Path:
	"""
	Converts a pdf file into an svg.

	Parameters
	----------
	input_file : str or pathlib.Path
		pdf file.

	Returns
	-------
	out: pathlib.Path
		The path to the svg file.

	"""

	input_file = pathlib.Path(input_file)
	output_file = input_file.with_suffix('.svg')

	path_to_pdf2svg = shutil.which('pdf2svg')

	assert path_to_pdf2svg, "couldn't find pdf2svg"

	command = [path_to_pdf2svg, input_file.name, output_file.name]

	run_summary = subprocess.run(command, capture_output=True, cwd=input_file.parent)

	assert run_summary.returncode == 0,\
		f"{colors.error}could not convert {colors.reset}{input_file}{colors.error} to svg"

	return output_file


def svg_to_html(input_file: Union[str, pathlib.Path]) -> str:
	"""

	Parameters
	----------
	input_file : str or pathlib.Path
		svg file.

	Returns
	-------

	"""

	# <!--?xml version\="1.0" encoding\="UTF-8"?-->

	# input file is read
	with open(input_file) as f:

		file_content = f.read()

	for to_be_escaped in [':', '~', '=', '#', '{', '}']:

		file_content = file_content.replace(to_be_escaped, '\\' + to_be_escaped)

	return r'<body>' + '\n' + file_content + r'</body>'



import sys
import pathlib
from typing import Union

import paramiko

from . import colors


class Connection:

	connection_not_available_help = (
		r'(you can try running the program in local mode, by passing "-l", or embedding the images, with "-e")')

	def __init__(self, host: str, user: str, password: str, public_key: Union[str, pathlib.Path]):

		self.host = host
		self.user = user
		self.password = password
		self.public_key = public_key

		# to be set in `connect`
		self.sftp = None

		# useful in `__del__` in the case the connection never gets established
		self.connection = None

		# self.connect()

	def connect(self):

		assert (self.password is not None) ^ (self.public_key is not None),\
			f'either "password" or "public_key" must be passed, but not both'

		if self.public_key is not None:

			# just in case "~" is in the given path
			public_key = pathlib.Path(self.public_key).expanduser()

			assert public_key.exists(), f'public key file, {public_key}, does not exist'

			# below, an actual string is needed
			public_key = public_key.as_posix()

		self.connection = paramiko.SSHClient()

		# so that it finds the key (no known_hosts error?)
		self.connection.set_missing_host_key_policy(paramiko.AutoAddPolicy())

		try:

			# connection is established
			self.connection.connect(self.host, username=self.user, password=self.password, key_filename=public_key)

		except paramiko.ssh_exception.AuthenticationException:

			# first character is not visible due to tqdm
			print(
				f'\n{colors.error}provided username {colors.reset}({self.user}){colors.error}'
				f' and/or password are not valid {self.connection_not_available_help}')

			sys.exit(1)

		except paramiko.ssh_exception.SSHException:

			# first character is not visible due to tqdm
			print(
				f'\n{colors.error}the provided public key {colors.reset}({self.public_key}){colors.error}'
				f' is not valid or has not been decrypted {self.connection_not_available_help}')

			sys.exit(1)

		# FTP component of the connection
		self.sftp = paramiko.SFTPClient.from_transport(self.connection.get_transport())

	def __del__(self):

		if self.connection is not None:

			self.connection.close()

	def is_active(self):

		if self.connection is None:

			return False

		else:

			return self.connection.get_transport().is_active()

	def copy(self, source: Union[str, pathlib.Path], remote_directory: str):

		if self.connection is None:

			self.connect()

		local = pathlib.Path(source)
		remote_directory = pathlib.Path(remote_directory)

		assert local.exists(), f'file {local} does not exist'

		self.make_directory_at(remote_directory.relative_to(remote_directory.parts[0]), remote_directory.parts[0])

		remote = remote_directory / local.name

		self.sftp.put(local.as_posix(), self.sftp.normalize(remote.as_posix()))

	def make_directory_at(self, new: Union[str, pathlib.Path], at: str):

		if self.connection is None:

			self.connect()

		self.sftp.chdir(at)

		# for every path component in the `new` directory...
		for subdirectory in pathlib.Path(new).parts:

			# if the subdirectory does not exist...
			if subdirectory not in self.sftp.listdir('.'):

				# ...it is made...
				self.sftp.mkdir(subdirectory)

				# ...and becomes the current working directory
			self.sftp.chdir(subdirectory)

		# the "current working directory" in unset
		self.sftp.chdir(None)


class FakeConnection:
	"""
	For offline runs.
	"""

	def __init__(self, host: str) -> None:

		self.host = host
		self.already_copied = set()
		self.files_to_copy = []

	@staticmethod
	def is_active():

		return False

	def copy(self, source: Union[str, pathlib.Path], remote_directory: str):

		source = pathlib.Path(source)

		if source.as_posix() not in self.already_copied:

			self.already_copied.add(source.as_posix())
			self.files_to_copy.append((source, remote_directory))

	@staticmethod
	def make_directory_at(new: str, at: str):

		pass



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gift-wrapper", # Replace with your own username
    version="1.6.1",
    author="Manuel A. Vázquez",
    author_email="manuavazquez@gmail.com",
    description="Build GIFT (Moodle compatible) files easily",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/manuvazquez/gift-wrapper",
    packages=setuptools.find_packages(),
    install_requires=['paramiko>=2.7.1', 'colorama>=0.4.3', 'PyYAML>=5.3.1', 'tqdm>=4.44.1'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': ['wrap.py=gift_wrapper.core:main'],
    }
)
